module FlexArray (Flex(Null,Leaf,Fork), empty, size, isEmpty, access, update, hiextBad, hiext, hirem, loext, lorem, toList) where

data Flex a = Null | Leaf a | Fork Int (Flex a) (Flex a)

-- invariante:
-- baum-struktur: leaf-tree
isLeafTree :: Flex a -> Bool
isLeafTree xt = isEmpty xt || isLeafy xt

-- keine Null-Knoten
isLeafy :: Flex a -> Bool
isLeafy Null           = False
isLeafy (Leaf x)       = True
isLeafy (Fork n xt yt) = isLeafy xt && isLeafy yt


empty :: Flex a
empty = Null

isEmpty :: Flex a -> Bool
isEmpty Null           = True
isEmpty (Leaf x)       = False
isEmpty (Fork n xt yt) = False


size :: Flex a -> Int
size Null           = 0
size (Leaf x)       = 1
size (Fork n xt yt) = n

height :: Flex a -> Int
height Null           = 0
height (Leaf x)       = 0
height (Fork n xt yt) = 1 + (height xt `max` height yt)


-- O(h)
access :: Flex a -> Int -> a
access (Leaf x) 0 = x
access (Fork n xt yt) k
	| k < m     = access xt k
	| otherwise = access yt (k - m)
	where m = size xt

-- O(h)
update :: Flex a -> Int -> a -> Flex a
update (Leaf y) 0 x = Leaf x
update (Fork n xt yt) k x
	| k < m     = Fork n (update xt k x) yt
	| otherwise = Fork n xt (update yt k x)
	where m = size xt


-- erhoeht height immer um 1
hiextBad :: a -> Flex a -> Flex a
hiextBad x Null           = Leaf x
hiextBad x (Leaf y)       = Fork 2 (Leaf y) (Leaf x)
hiextBad x (Fork n xt yt) = Fork (n+1) (Fork n xt yt) (Leaf x)

-- minimale height
hiext :: a -> Flex a -> Flex a
hiext x Null           = Leaf x
hiext x (Leaf y)       = Fork 2 (Leaf y) (Leaf x)  -- neue fork, rechts einhaengen

hiext x (Fork n xt yt) = Fork (n + 1) xt (hi x yt) -- in den rechten subtree einhaengen
	where
		hi x (Leaf y) = Fork 2 (Leaf y) (Leaf x) -- neue fork, rechts einhaengen
		hi x (Fork n xt yt)
			-- rechts neben die fork haengen
			| size xt == size yt = Fork (n+1) (Fork n xt yt) (Leaf x)
		
			-- links beibehalten, in den rechten subtree einhaengen
			| otherwise          = Fork (n+1) xt (hiext x yt) 


loext :: a -> Flex a -> Flex a
loext x Null = Leaf x
loext x (Leaf y) = Fork 2 (Leaf x) (Leaf y) -- neue fork, links einhaengen

loext x (Fork n xt yt) = Fork (n+1) (lo x xt) yt -- in den linken subtree einhaengen 
	where
		lo x (Leaf y) = Fork 2 (Leaf x) (Leaf y)
		lo x (Fork n xt yt)
			-- links neben die fork haengen
			| size xt == size yt = Fork (n+1) (Leaf x) (Fork n xt yt)

			-- rechts beibehalten, in den linken subtree einhaengen
			| otherwise          = Fork (n+1) (lo x xt) yt


hirem :: Flex a -> Flex a
hirem (Leaf x)           = Null
hirem (Fork (n+1) xt yt) = fork n xt (hirem yt) -- rechts rauschloeschen


lorem :: Flex a -> Flex a
lorem (Leaf x) = Null
lorem (Fork n xt yt) = fork (n-1) (lorem xt) yt -- links rausloeschen


fork :: Int -> Flex a -> Flex a -> Flex a
fork n xt yt
	| isEmpty xt = yt
	| isEmpty yt = xt
	| otherwise  = Fork n xt yt


toList :: Flex a -> [a]
toList Null           = []
toList (Leaf x)       = [x]
toList (Fork n xt yt) = toList xt ++ toList yt

instance Show a => Show (Flex a) where
	show (Leaf x) = show x
	show (Fork n xt yt) =  "<" ++ show xt ++ " [" ++ show n ++ "] " ++ show yt ++ ">"