{-
Repraesentation eines Sets als Baum
Quelle: Richard Bird - Introduction to Functional Programming using Haskell

-}

module TSet (Set, empty, isEmpty, member, insert, delete)
where
  data Set a = Null | 
                Fork (Set a) a (Set a)
                
  --leere Menge
  empty :: Set a
  empty = Null
  
  --Test auf leere Menge
  isEmpty :: Set a -> Bool
  isEmpty Null           = True
  isEmpty (Fork xt y zt) = False  
  
  --Test, ob Element in Menge ist
  member :: (Ord a) => Set a -> a -> Bool
  member Null x = False
  member (Fork xt y zt) x
    |(x < y)  = member xt x
    |(x == y) = True
    |(x > y)  = member zt x
  
  --Einfuegen eines Elements in eine Menge
  insert :: (Ord a) => a -> Set a -> Set a
  insert x Null = (Fork Null x Null)
  insert x (Fork xt y zt)
    |(x < y)  = Fork (insert x xt) y zt
    |(x == y) = Fork xt y zt
    |(x > y)  = Fork xt y (insert x zt)
  
  --loeschen eines Elements aus einer Menge
  delete :: (Ord a) => a -> Set a -> Set a
  delete x Null = Null
  delete x (Fork xt y zt)
    |(x < y)  = Fork (delete x xt) y zt
    |(x == y) = join xt zt
    |(x > y)  = Fork xt y (delete x zt)
    
    
      --Hilfsfunktionen
  join :: Set a -> Set a -> Set a
  join xt yt = if isEmpty yt then xt
               else Fork xt y zt
               where
                 (y,zt) = splitTree yt
    
  splitTree :: Set a -> (a, Set a)
  splitTree (Fork xt y zt) = if isEmpty xt then (y, zt)
                             else (u, Fork vt y zt)
                             where
                               (u, vt) = splitTree xt
                               
  reprn :: Ord a => Set a -> [a]
  reprn Null = []
  reprn (Fork xt y zt) = qsort( reprn xt ++ [y] ++ reprn zt)
  
  qsort :: Ord a => [a] -> [a]
  qsort []     = []
  qsort (x:xs) = qsort kleiner ++ [x] ++ qsort groesser
                 where
                    kleiner  = [y | y <- xs, y < x]
                    groesser = [y | y <- xs, y >= x]
  
  instance Show a => Show (Set a)
    where
    show Null         =  "-"
    show (Fork l a r) =  "(Fork "
                    ++ show l ++ " " 
                    ++ show a ++ " " 
                    ++ show r ++ ")"   
                    
  instance Ord a => Eq (Set a)
    where
    x == y = reprn x == reprn y
  
